"""
.. module:: Imputer

Imputer
*************

:Description: KnnImputer

    Class for the imputation of missing values using the k nearest neigbours


:Authors: bejar
    

:Version: 

:Created on: 01/09/2015 14:17 

"""

__author__ = 'bejar'


import numpy as np
from sklearn.base import TransformerMixin
from sklearn.neighbors import NearestNeighbors

class KnnImputer(TransformerMixin):
    """
    Missing values imputation using the mean of the k-neighbors considering the
    dimensions that are not missing.

    It only uses the examples that do not have any missing value

    Parameters:

    missing_values: float or 'NaN'
     Value that indicates a missing value

    n_neighbors: int
       The number of neighbors to consider
    distance: str
       distance to use to compute the neighbors ('euclidean')
    """
    neigh = None
    miss_val = None
    dist = None
    miss_ind_ = None

    def __init__(self, missing_values='NaN', n_neighbors=1, distance='euclidean'):
        self.neigh = n_neighbors
        self.miss_val = float(missing_values)
        self.dist = distance

    def fit(self):
        """
        does nothing

        """

    def _transform(self, X):
        """
        Imputes the missings
        :param X:
        :return:
        """

        l_miss_ex = []
        l_no_miss_ex = []
        self.miss_ind_ = []
        for row in range(X.shape[0]):
            l_miss_att = []
            for column in range(X.shape[1]):
                if np.isnan(X[row, column]) or  X[row, column] ==self.miss_val:
                    l_miss_att.append(column)

            if l_miss_att:
                l_miss_ex.append((row, l_miss_att))
                self.miss_ind_.append(row)
            else:
                l_no_miss_ex.append(row)

        if not l_no_miss_ex:
            raise Exception('KnnImputer: All examples have missing values')
        else:
            nomiss = X[l_no_miss_ex]
            if nomiss.shape[0] < self.neigh:
                raise Exception('KnnImputer: Not enough examples without missings')
            for ex, att in l_miss_ex:
                l_sel = [s for s in range(X.shape[1]) if s not in att]
                knn = NearestNeighbors(n_neighbors=self.neigh, metric=self.dist)
                knn.fit(nomiss[:, l_sel])

                l_neigh = knn.kneighbors(X[ex][l_sel], return_distance=False)[0]
                for a in att:
                    l_mean = nomiss[l_neigh, a]
                    X[ex][a] = np.mean(l_mean)
        return X

    def fit_transform(self, X, copy=True):
        """
        Looks for the examples with missing values and computes the new values

        :param matrix X: data matrix
        :param bool copy: If True returns a copy of the data
        :return:
        """
        if copy:
            y = X.copy()
        else:
            y = X

        self._transform(y)

        return y


if __name__ == '__main__':

    mean, cov = [0, 0, 0], [(1, .5, .5), (.5, 1, .5), (.5, .5, 1)]
    data = np.random.multivariate_normal(mean, cov, 200)
    vals = np.random.choice(200, size=20, replace=False)

    for v in vals[0:20]:
        data[v][0] = np.nan

    kimp = KnnImputer(n_neighbors=2)

    data2 = kimp.fit_transform(data)

    print (kimp.miss_ind_)
    for i in range(data.shape[0]):
        print (data[i], data2[i])


"""
.. module:: Leader

Leader
*************

:Description: Leader Algorithm Clustering

:Authors: Javier Bejar

:Version: 1.0

:Created on: 07/07/2014 8:29

"""

__author__ = 'bejar'

import numpy as np
from sklearn.base import BaseEstimator, ClusterMixin, TransformerMixin
from sklearn.metrics.pairwise import euclidean_distances


class Leader(BaseEstimator, ClusterMixin, TransformerMixin):
    """Leader Algorithm Clustering

    :param float radius: Clustering radius for assigning examples to a cluster

    """
    def __init__(self, radius):
        self.radius = radius
        self.cluster_centers_ = None
        self.labels_ = None
        self.cluster_sizes_ = None

    def num_clusters(self):
        return self.cluster_centers_.shape[0]

    def fit(self,X):
        """Clusters the examples

        :param matrix X: Data

        """

        self.cluster_centers_, self.labels_, self.cluster_sizes_ = self._fit_process(X)

        return self

    def predict(self, X):
        """Returns the nearest cluster for a data matrix

        :param matrix X: Data
        :return: Nearest clusters for the data
        """
        clasif = []
        for i in range(X.shape[0]):
            ncl, mdist = self._find_nearest_cluster(X[i], self.cluster_centers_)
            if mdist <= self.radius:
                clasif.append(ncl)
            else:
                clasif.append(-1)
        return clasif

    def _fit_process(self, X):
        """
        Clusters incrementally the examples

        :param X:
        :return:
        """
        assignments = []
        scenters = np.zeros((1,X.shape[1]))
        centers = np.zeros((1,X.shape[1]))
        # Initialize with the first example
        scenters[0] = X[0]
        centers[0] = X[0]
        assignments.append([0])
        csizes = np.array([1])
        # Cluster the rest of examples
        for i in range(1,X.shape[0]):
            ncl,mdist = self._find_nearest_cluster(X[i], centers)

            # if distance is less than radius, introduce example in nearest class
            if mdist <= self.radius:
                scenters[ncl] += X[i]
                csizes[ncl] += 1
                centers[ncl] = scenters[ncl]/csizes[ncl]
                assignments[ncl].append(i)
            else: # Create a new cluster
                scenters = np.append(scenters,np.array([X[i]]), 0)
                centers = np.append(centers,np.array([X[i]]), 0)
                csizes = np.append(csizes, [1], 0)
                assignments.append([i])

        labels = np.zeros(X.shape[0])
        for l, ej in enumerate(assignments):
            for e in ej:
                labels[e] = l

        return centers, labels, csizes

    @staticmethod
    def _find_nearest_cluster(examp, centers):
        """
        Finds the nearest cluster for an example

        :param examp:
        :param centers:
        :return:
        """

        dist = euclidean_distances(centers, examp)

        pmin = np.argmin(dist)
        vmin = np.min(dist)

        return pmin, vmin
